/*
        gc.c -- C Generic Container 2.0
        (C) Copyright 1996  John Webster Small
        All rights reserved
*/

#include "gc.h"

/*lint -e613 */

void * gc_memdup(const void * D, size_t size)
{
    void * cD = 0;

    if (D && size)
        if ((cD = GC_NEW_DATA(size)) != (void *)0)
            (void) memcpy(cD,D,size);
    return cD;
}

#pragma argsused
/* ARGSUSED */
static gc_vf_assignD(_gc_string)
{ return strncpy(D,S,strlen(D)+1); }

#pragma argsused
/* ARGSUSED */
static gc_vf_newD(_gc_string)
{ return (_gc_string *) gc_memdup(D,strlen(D)+1); }

gc_vft_def(_gc_string,
    (gc_vF_assignD)(gc_vf_name(_gc_string,assignD)),
    (gc_vF_newD)(gc_vf_name(_gc_string,newD)),
    (gc_vF_deleteD)(0),
    (gc_vF_attachD)(0),
    (gc_vF_detachD)(0));

gc_def(int);
gc_def(long);
gc_def(short);
gc_def(unsigned);
gc_def(_gc_ulong);
gc_def(_gc_ushort);
gc_def(float);
gc_def(double);
gc_def(_gc_ldouble);

gc * _gc_init_specs(gc * this_gc, gc_vft * vfT,
    unsigned flags_, unsigned short dynamic_)
{
    assert(this_gc);
    this_gc->_curNode =  this_gc->first = 0U;
    this_gc->_nodes = 0U;
    this_gc->_cmP = voidCmP0;
    this_gc->vfT = vfT;
  /*
    The following relationships are maintained
    during operation of a gc:
    1 <= delta <= lowLimit <= limit <= maxNodes
           <= GC_MAXNODES
    lowThreshold = lowLimit - delta;
  */
    assert(((unsigned)GC_DELTA));
    assert(((unsigned)GC_DELTA) <= ((unsigned)GC_LIMIT));
    assert(((unsigned)GC_LIMIT) <= ((unsigned)GC_MAXNODES));
    this_gc->linkS = (void **)0;
    this_gc->_limit = ((unsigned)GC_LIMIT);
    this_gc->_delta = ((unsigned)GC_DELTA);
    this_gc->_maxNodes = ((unsigned)GC_MAXNODES);
    this_gc->lowLimit = ((unsigned)GC_LIMIT);
    this_gc->lowThreshold =
        ((unsigned)GC_LIMIT) - ((unsigned)GC_DELTA);
    this_gc->_flags = (flags_ | GC_SORTED);
    this_gc->dynamic = dynamic_;
    return this_gc;
}

gc * gc_init_specs(gc * this_gc_0, gc_vft * vfT,
    unsigned flags_)
{
    if (!this_gc_0)  {
        this_gc_0 = GC_NEW_HEADER;
        if (!this_gc_0)
            return (gc *)0;
        return _gc_init_specs(this_gc_0,vfT,flags_,1);
    }
    return _gc_init_specs(this_gc_0,vfT,flags_,0);
}

gc * gc_copyInit_specs
    (gc * this_gc_0, const gc * src_gc, int deep)
{
    unsigned i;

    _gc_assert_valid(src_gc);
    this_gc_0 = gc_init_specs(this_gc_0,src_gc->vfT,
        gc_flags_specs(src_gc,GC_ASG));
    if (!this_gc_0)
        return (gc *)0;
    if (deep)  {
      gc_setFlags(this_gc_0,GC_NEW | GC_DEL);
      for (i = 0U; i < gc_nodes(src_gc); i++)
          if (!gc_insQNew(this_gc_0,gc_atGet(src_gc,i)))
              break;
    }
    else  {
      for (i = 0U; i < gc_nodes(src_gc); i++)
          if (!gc_insQ(this_gc_0,gc_atGet(src_gc,i)))
              break;
    }
    gc_setCmP(this_gc_0,gc_cmP(src_gc));
    gc_setFlags(this_gc_0,gc_flags_specs(src_gc,GC_SORTED));
    (void) gc_setCurNode(this_gc_0,gc_curNode(src_gc));
    return this_gc_0;
}

gc * gc_destruct(gc * this_gc)
{
    _gc_assert_valid(this_gc);
    gc_allClr(this_gc);
    this_gc->linkS = (void **) this_gc;
    if (this_gc->dynamic)  {
        GC_DELETE_HEADER(this_gc);
        return (gc *)0;
    }
    return this_gc;
}

int gc_cmp(const gc * this_gc, const gc * b)
{
    int signum;
    unsigned i;

    _gc_assert_valid(this_gc);
    _gc_assert(this_gc->_cmP && b);
    for (i = 0;
        i < gc_nodes(this_gc) && i < gc_nodes(b);
        i++)
    {
        signum = (*(this_gc->_cmP))
            (gc_atGet(this_gc,i),gc_atGet(b,i));
        if (signum)
           return signum;
    }
    if (this_gc->_nodes > b->_nodes)
        return 1;
    if (this_gc->_nodes < b->_nodes)
        return -1;
    return 0;
}

#ifndef GC_NDEBUG
gc * _gc_valid(const gc * this_gc)
{
    _gc_assert_valid(this_gc);
    return (gc *) this_gc;
}
#endif

void * _gc_atDeleteLeak(gc * this_gc, unsigned n, void * D,
    void * (*method)(gc *, unsigned, void *))
{
    if (D)  {
        void * ok = (*method)(this_gc,n,D);
        if (!ok)
            if (this_gc->vfT && this_gc->vfT->deleteD)
                (*(this_gc->vfT->deleteD))(this_gc,D);
            else
                GC_DELETE_DATA(D);
        return ok;
    }
    return 0;
}

void * _gc_deleteLeak
    (gc * this_gc, void * D, void * (*method)(gc *, void *))
{
    if (D)  {
        void * ok = (*method)(this_gc,D);
        if (!ok)
            if (this_gc->vfT && this_gc->vfT->deleteD)
                (*(this_gc->vfT->deleteD))(this_gc,D);
            else
                GC_DELETE_DATA(D);
        return ok;
    }
    return 0;
}

void * _gc_cloneD(gc * this_gc, const void  * D)
{
    _gc_assert_valid(this_gc);
    if (D && (this_gc->_flags & GC_NEW) &&
        this_gc->vfT && this_gc->vfT->newD)
        return (*(this_gc->vfT->newD))(this_gc,D);
    return 0;
}

void gc_setLimit(gc * this_gc, unsigned newLimit)
{
    void ** newLinkS;
    unsigned headNodes;

    _gc_assert_valid(this_gc);
    if (newLimit < this_gc->_nodes)
        newLimit = this_gc->_nodes;
    else if (newLimit > this_gc->_maxNodes)
        newLimit = this_gc->_maxNodes;
    if (newLimit < this_gc->_delta)
        newLimit = this_gc->_delta;
    if (!newLimit || newLimit == this_gc->_limit)
        return;
    if (!this_gc->linkS)  {
        this_gc->_limit = newLimit;
        return;
    }
    newLinkS = GC_NEW_VECTOR(newLimit);
    if (!newLinkS)
        return;
    if ((headNodes = this_gc->_limit - this_gc->first) > this_gc->_nodes)
        headNodes = this_gc->_nodes;
    if (headNodes)  /* move leading nodes */
        memcpy(newLinkS,&this_gc->linkS[this_gc->first],
           sizeof(void*)*headNodes);
    /* copy wrap around trailing nodes */
    if (headNodes < this_gc->_nodes)
        memcpy(&newLinkS[headNodes],this_gc->linkS,
           sizeof(void*)*(this_gc->_nodes-headNodes));
    if (newLimit > this_gc->_limit)
        if (newLimit - this_gc->_delta > this_gc->_limit)
            this_gc->lowLimit = newLimit - this_gc->_delta;
        else
            this_gc->lowLimit = this_gc->_limit;
    else
        if (newLimit - this_gc->_delta > this_gc->_delta)
            this_gc->lowLimit = newLimit - this_gc->_delta;
        else
            this_gc->lowLimit = this_gc->_delta;
    this_gc->lowThreshold = this_gc->lowLimit - this_gc->_delta;
    GC_DELETE_VECTOR(this_gc->linkS);
    this_gc->linkS = newLinkS;
    this_gc->_limit = newLimit;
    this_gc->first = 0U;
    return;
}

void gc_setDelta(gc * this_gc, unsigned newDelta)
{
    _gc_assert_valid(this_gc);
    if (newDelta && (newDelta <= this_gc->lowLimit))
        this_gc->_delta = newDelta;
}

void gc_setMaxNodes(gc * this_gc, unsigned newMaxNodes)
{
    _gc_assert_valid(this_gc);
    if (newMaxNodes < this_gc->_limit)
        if (newMaxNodes > this_gc->_nodes)
            gc_setLimit(this_gc,newMaxNodes);
    if (newMaxNodes >= this_gc->_limit)
        if (newMaxNodes > GC_MAXNODES)
            this_gc->_maxNodes = GC_MAXNODES;
        else
            this_gc->_maxNodes = newMaxNodes;
}

void * gc_atIns(gc * this_gc, unsigned n, void * D)
{
    void ** newLinkS;
    unsigned newLimit, i;
    int ok;

    _gc_assert_valid(this_gc);
    if (!D)
        return (void *)0;
    if (this_gc->_nodes == this_gc->_limit)  { /* no vacancy - expand */
        if (this_gc->_limit == this_gc->_maxNodes)
            return (void *)0;
        newLimit = ((this_gc->_maxNodes - this_gc->_delta > this_gc->_limit)?
            (this_gc->_limit + this_gc->_delta) : this_gc->_maxNodes);
        newLinkS = GC_NEW_VECTOR(newLimit);
        if (!newLinkS)  /* expansion unavailable */
            return (void *)0;
        if (this_gc->vfT && this_gc->vfT->attachD)
            ok = (*(this_gc->vfT->attachD))(this_gc,D);
        else
            ok = 1;
        if (!ok)  { /* data refuses binding */
            GC_DELETE_VECTOR(newLinkS);
            return (void *)0;
        }
        newLinkS[((n <= this_gc->_nodes)? n : (n = this_gc->_nodes))] = D;
        /* calc headNodes */
        if (n <= (i = this_gc->_limit - this_gc->first))  {
            /* insert in head section */
            if (n)  { /* postfix or split head */
                memcpy(newLinkS,&this_gc->linkS[this_gc->first],
                    sizeof(void*)*n);
                if (n < i) /* it's a split */
                    memcpy(&newLinkS[n+1],
                        &this_gc->linkS[this_gc->first+n],
                        sizeof(void*)*(i-n));
            }
            else    /* prefix head */
                memcpy(&newLinkS[1],&this_gc->linkS[this_gc->first],
                    sizeof(void*)*i);
            if (i < this_gc->_nodes)  /* copy tail */
                memcpy(&newLinkS[1+i],this_gc->linkS,
                    sizeof(void*)
                    *(this_gc->_nodes-i));
        }
        else  {    /* insert in tail section */
            /* copy head this_gc->first */
            memcpy(newLinkS,&this_gc->linkS[this_gc->first],
                sizeof(void*)*i);
            if (i < this_gc->_nodes)  { /* copy tail */
                memcpy(&newLinkS[i],this_gc->linkS,
                    sizeof(void*)*(n-i));
                if (n < this_gc->_nodes) /* it's a tail split */
                    memcpy(&newLinkS[n+1],
                        &this_gc->linkS[n-i],
                        sizeof(void*)
                        *(this_gc->_nodes-n));
            }
        }
        /* Compute next smaller this_gc->linkS size */
        /* and threshold for shrinking. */
        this_gc->lowLimit = this_gc->_limit;
        this_gc->lowThreshold = this_gc->lowLimit - this_gc->_delta;
        /* swap new for old */
        GC_DELETE_VECTOR(this_gc->linkS);
        this_gc->linkS = newLinkS;
        this_gc->_limit = newLimit;
        this_gc->first = 0U;
    }
    else  {  /* vacancy */
        if (!this_gc->linkS)
            this_gc->linkS = GC_NEW_VECTOR(this_gc->_limit);
        if (!this_gc->linkS)
                return (void *)0;
        if (this_gc->vfT && this_gc->vfT->attachD)
            if (!(*(this_gc->vfT->attachD))(this_gc,D))  {
                if (!this_gc->_nodes)  {
                    GC_DELETE_VECTOR(this_gc->linkS);
                    this_gc->linkS = (void **)0;
                }
                return (void *)0;
            }
        if (!n)  /*  push */
            this_gc->linkS[this_gc->first? --this_gc->first
                : (this_gc->first = this_gc->_limit - 1)] = D;
        else if (n >= this_gc->_nodes) /*  insQ */
            this_gc->linkS[(this_gc->first
               +(n=this_gc->_nodes))%this_gc->_limit] = D;
        else  { /*  insert interior */
            i = (this_gc->first + n) % this_gc->_limit;
            if (i < this_gc->first || !this_gc->first)
                /*  move rear rightward */
                memmove(&this_gc->linkS[i+1],&this_gc->linkS[i],
                    sizeof(void*)
                    * (this_gc->_nodes-n));
            else  { /*  move front leftward */
                memmove(&this_gc->linkS[this_gc->first-1],
                    &this_gc->linkS[this_gc->first],
                    sizeof(void*)
                    *(n));
                this_gc->first--;
                i--;
            }
            this_gc->linkS[i] = D;
        }
    }
    this_gc->_nodes++;
    if (n <= this_gc->_curNode)
        this_gc->_curNode++;
    this_gc->_flags &= ~GC_SORTED;
    return D;
}

void * gc_atRmv(gc * this_gc, unsigned n)
{
    void ** newLinkS = (void **)0;
    unsigned newLimit, i;
    void * D;

    _gc_assert_valid(this_gc);
    if (!this_gc->linkS || n >= this_gc->_nodes)
        return (void *)0;
    D = this_gc->linkS[i=((this_gc->first+n)%this_gc->_limit)];
    if (this_gc->_nodes <= this_gc->lowThreshold) {
        /*  get ready to contract */
        newLimit = this_gc->lowLimit;
        newLinkS = GC_NEW_VECTOR(newLimit);
        /*  Compute next smaller this_gc->linkS */
        /*  size and threshold for */
        /*  shrinking. */
        if (newLinkS)  {
            if (this_gc->_nodes < this_gc->_delta)
                this_gc->lowThreshold = 0;
            else
                this_gc->lowThreshold -= this_gc->_delta;
            this_gc->lowLimit = this_gc->lowThreshold + this_gc->_delta;
        }
    }
    if (newLinkS)  {  /*  remove and contract */
        /*  calc head nodes */
        if ((i = this_gc->_limit - this_gc->first) > this_gc->_nodes)
                i = this_gc->_nodes;
        if (n < i)  {
            /*  rmv from head section */
            if (n)  { /*  rmv from head */
                memcpy(newLinkS,&this_gc->linkS[this_gc->first],
                    sizeof(void*)*n);
                if (n < i - 1)
                        memcpy(&newLinkS[n],
                            &this_gc->linkS[this_gc->first+n+1],
                            sizeof(void*)*(i-n-1));
            }
            else if (i > 1)   /*  pop head */
                memcpy(newLinkS,&this_gc->linkS[this_gc->first+1],
                    sizeof(void*)*(i-1));
            if (i < this_gc->_nodes)  /*  copy tail */
                memcpy(&newLinkS[i-1],this_gc->linkS,
                    sizeof(void*)
                    *(this_gc->_nodes-i));
        }
        else  { /*  i <= n < nodes */
            /*  rmv from tail section */
            /*  copy head this_gc->first */
            memcpy(newLinkS,&this_gc->linkS[this_gc->first],
                sizeof(void*)*i);
            if (i == n)  { /*  pop tail */
                if (n < this_gc->_nodes - 1)
                    /*  copy trailing tail */
                    memcpy(&newLinkS[i],&this_gc->linkS[1],
                        sizeof(void*)*(this_gc->_nodes-n-1));
            }
            else  {  /*  copy up to n in tail */
                memcpy(&newLinkS[i],this_gc->linkS,
                    sizeof(void*)*(n-i));
                /*  skip n'th node */
                if (n < this_gc->_nodes - 1)
                    /*  copy anything trailing n'th node */
                    memcpy(&newLinkS[n],&this_gc->linkS[n-i+1],
                        sizeof(void*)*(this_gc->_nodes-n-1));
            }
        }
        /*  swap new for old */
        GC_DELETE_VECTOR(this_gc->linkS);
        this_gc->linkS = newLinkS;
        this_gc->_limit = newLimit;    /*lint !e644 */
        this_gc->first = 0U;
    }
    else  { /*  simply remove */
        if (!n)  {  /*  pop */
            if (++this_gc->first >= this_gc->_limit)
                this_gc->first = 0U;
        }
        else if (n != this_gc->_nodes-1)  /*  del interior */
            if (i < this_gc->first) /*  move tail leftward */
                memmove(&this_gc->linkS[i],&this_gc->linkS[i+1],
                    sizeof(void*)
                    *(this_gc->_nodes-n-1));
            else {  /*  move head rightward */
                memmove(&this_gc->linkS[this_gc->first+1],
                    &this_gc->linkS[this_gc->first],
                    sizeof(void*)*n);
                this_gc->first++;
            }
    }
    if (n < this_gc->_curNode)
        this_gc->_curNode--;
    else if (n == this_gc->_curNode)
        this_gc->_curNode = this_gc->_nodes-1;
    if (--this_gc->_nodes == 0U)  {
        this_gc->_flags |= GC_SORTED;
        this_gc->_flags &= ~GC_NEWED;
        GC_DELETE_VECTOR(this_gc->linkS);
        this_gc->linkS = (void **)0;
    }
    if (this_gc->vfT && this_gc->vfT->detachD)
        (*(this_gc->vfT->detachD))(this_gc,D);
    return D;
}

void gc_allRmv(gc * this_gc)
{
    unsigned i, j;

    _gc_assert_valid(this_gc);
    if (this_gc->linkS)  {
        if (this_gc->vfT && this_gc->vfT->detachD)  {
            for (i = this_gc->first, j = this_gc->_nodes;
              i < this_gc->_limit && j; i++, j--)
              (*(this_gc->vfT->detachD))(this_gc,this_gc->linkS[i]);
            for (i = 0; j; i++, j--)
              (*(this_gc->vfT->detachD))(this_gc,this_gc->linkS[i]);
        }
        GC_DELETE_VECTOR(this_gc->linkS);
    }
    (void) _gc_init_specs(this_gc,this_gc->vfT,
        this_gc->_flags,this_gc->dynamic);
}

int  gc_atDel(gc * this_gc, unsigned n)
{
    void * D;

    _gc_assert_valid(this_gc);
    if (this_gc->_flags & GC_DEL)
        if ((D = gc_atRmv(this_gc,n)) != (void *)0)  {
            if (this_gc->vfT && this_gc->vfT->deleteD)
                (*(this_gc->vfT->deleteD))(this_gc,D);
            else
                GC_DELETE_DATA(D);
            return 1;
        }
    return 0;
}

int gc_allDel(gc * this_gc)
{
    _gc_assert_valid(this_gc);
    if (this_gc->_flags & GC_DEL)  {
        if (this_gc->linkS)  {
            unsigned i = this_gc->first, j = this_gc->_nodes;
            for (i = this_gc->first, j = this_gc->_nodes;
                i < this_gc->_limit && j; i++, j--)  {
                if (this_gc->vfT)  {
                    if (this_gc->vfT->detachD)
                        (*(this_gc->vfT->detachD))(this_gc,this_gc->linkS[i]);
                    if (this_gc->vfT->deleteD)
                        (*(this_gc->vfT->deleteD))(this_gc,this_gc->linkS[i]);
                    else
                        GC_DELETE_DATA(this_gc->linkS[i]);
                }
                else
                    GC_DELETE_DATA(this_gc->linkS[i]);
            }
            for (i = 0; j; i++, j--)  {
                if (this_gc->vfT)  {
                    if (this_gc->vfT->detachD)
                        (*(this_gc->vfT->detachD))(this_gc,this_gc->linkS[i]);
                    if (this_gc->vfT->deleteD)
                        (*(this_gc->vfT->deleteD))(this_gc,this_gc->linkS[i]);
                    else
                        GC_DELETE_DATA(this_gc->linkS[i]);
                }
                else
                    GC_DELETE_DATA(this_gc->linkS[i]);
            }
            GC_DELETE_VECTOR(this_gc->linkS);
        }
        (void) _gc_init_specs(this_gc,this_gc->vfT,
            this_gc->_flags,this_gc->dynamic);
        return 1;
    }
    return 0;
}

void gc_allClr(gc * this_gc)
{
    _gc_assert_valid(this_gc);
    if (this_gc->_flags & GC_DEL)
        (void) gc_allDel(this_gc);
    else
        gc_allRmv(this_gc);
}

void * gc_atPut(gc * this_gc, unsigned n, void * D)
{
    int ok;
    void ** N;

    _gc_assert_valid(this_gc);
    if (this_gc->linkS && D && (n < this_gc->_nodes))  {
        N = &this_gc->linkS[(this_gc->first+n)%this_gc->_limit];
        if (D != *N)  {
            if (this_gc->vfT && this_gc->vfT->attachD)
                ok = (*(this_gc->vfT->attachD))(this_gc,D);
            else
                ok = 1;
            if (ok)  {
                this_gc->_flags &= ~GC_SORTED;
                if (this_gc->vfT && this_gc->vfT->detachD)
                    (*(this_gc->vfT->detachD))(this_gc,*N);
                if (this_gc->_flags & GC_DEL)
                    if (this_gc->vfT && this_gc->vfT->deleteD)
                        (*(this_gc->vfT->deleteD))(this_gc,*N);
                    else
                        GC_DELETE_DATA(*N);
                return (*N=D);
            }
        }
    }
    return (void *)0;
}

void * gc_atPutNew(gc * this_gc, unsigned n, const void * D)
{
    void * oldD, * cD;

    _gc_assert_valid(this_gc);
    if (!D || !(this_gc->_flags & GC_NEW)
        || !(this_gc->vfT) || !(this_gc->vfT->newD))
        return (void *)0;
    if ((oldD = gc_atGet(this_gc,n)) == (void *)0)
        return (void *)0;
    if (oldD == D)
        return (void *)0;
    cD = (*(this_gc->vfT->newD))(this_gc,D);
    if (cD)  {
        if (gc_atPut(this_gc,n,cD))  {
            this_gc->_flags |= GC_NEWED;
            return cD;
        }
        else if (this_gc->vfT->deleteD)
            (*(this_gc->vfT->deleteD))(this_gc,cD);
        else
            GC_DELETE_DATA(cD);
    }
    return (void *)0;
}

void * gc_atPutAsg(gc * this_gc, unsigned n, const void * D)
{
    void * oldD;

    _gc_assert_valid(this_gc);
    if (D && this_gc->_flags & GC_ASG
        && this_gc->vfT && this_gc->vfT->assignD)
        if ((oldD = gc_atGet(this_gc,n)) != (void *)0)
            if (D == oldD)
                return oldD;
            else  {
                this_gc->_flags &= ~(GC_READ | GC_READ_DEL);
                if ((*(this_gc->vfT->assignD))
                    (this_gc,oldD,D,this_gc->_flags))
                    return oldD;
            }
    return (void *)0;
}

void * gc_atGet(const gc * this_gc, unsigned n)
{
    _gc_assert_valid(this_gc);
    return ((this_gc->linkS && (n < this_gc->_nodes))?
       this_gc->linkS[(this_gc->first+n)%this_gc->_limit] : (void *)0);
}

void * gc_atGetAsg(gc * this_gc, unsigned n, void * D)
{
    void * N;

    _gc_assert_valid(this_gc);
    if (D && this_gc->_flags & GC_ASG
        && this_gc->vfT && this_gc->vfT->assignD)
        if ((N = gc_atGet(this_gc,n)) != (void *)0)
            if (D == N)
                return D;
            else  {
                this_gc->_flags &= ~GC_READ_DEL;
                this_gc->_flags |=  GC_READ;
                if ((*(this_gc->vfT->assignD))
                    (this_gc,D,N,this_gc->_flags))
                    return D;
            }
    return (void *)0;
}

void * gc_atXchg(gc * this_gc, unsigned n, void * D)
{
    int ok;
    void ** N, * oldD;

    _gc_assert_valid(this_gc);
    if (this_gc->linkS && D && (n < this_gc->_nodes))  {
        if (this_gc->vfT && this_gc->vfT->attachD)
            ok = ((*(this_gc->vfT->attachD))(this_gc,D));
        else
            ok = 1;
        if (ok)  {
            this_gc->_flags &= ~GC_SORTED;
            N = &this_gc->linkS[(this_gc->first+n)
                 %this_gc->_limit];
            oldD = *N;
            *N = D;
            if (this_gc->vfT && this_gc->vfT->detachD)
                (*(this_gc->vfT->detachD))(this_gc,oldD);
            return oldD;
        }
    }
    return (void *)0;
}

unsigned gc_indexOf(const gc * this_gc, const void * D)
{
    unsigned i;

    _gc_assert_valid(this_gc);
    if (this_gc->linkS && D)
        for (i = 0U; i < this_gc->_nodes; i++)
            if (D == this_gc->linkS[(this_gc->first+i)
                %this_gc->_limit])
                return i;
    return GC_NOTFOUND;
}

void *  gc_setCurNode(gc * this_gc, unsigned n)
{
    _gc_assert_valid(this_gc);
    if (n > this_gc->_nodes)
        this_gc->_curNode = this_gc->_nodes;
    else
        this_gc->_curNode = n;
    if ((this_gc->_curNode < this_gc->_nodes) && this_gc->linkS)
        return this_gc->linkS
            [(this_gc->first+this_gc->_curNode)%this_gc->_limit];
    return (void *) 0;
}

void * gc_insBefore(gc * this_gc, void * D)
{ 
    unsigned n;

    _gc_assert_valid(this_gc);
    n = this_gc->_curNode;
    if (n > this_gc->_nodes)
        n = this_gc->_nodes;
    D =  gc_atIns(this_gc,n,D);
    this_gc->_curNode = n;
    return D;
}

void * gc_rmvStayPut(gc * this_gc)
{ 
    unsigned n;
    void * D;

    _gc_assert_valid(this_gc);
    n = this_gc->_curNode;
    D =  gc_atRmv(this_gc,n);
    this_gc->_curNode = n;
    return D;
}

int gc_delStayPut(gc * this_gc)
{
    unsigned n;
    int ok;

    _gc_assert_valid(this_gc);
    n = this_gc->_curNode;
    ok = gc_atDel(this_gc,n);
    this_gc->_curNode = n;
    return ok;
}

int gc_clrStayPut(gc * this_gc)
{
    _gc_assert_valid(this_gc);
    if (this_gc->_flags & GC_DEL)
        return gc_delStayPut(this_gc);
    return (gc_rmvStayPut(this_gc) != 0);
}

void * gc_insAfter(gc * this_gc, void * D)
{
    _gc_assert_valid(this_gc);
    if (gc_atIns(this_gc,this_gc->_curNode+1,D))  {
        if (++this_gc->_curNode >= this_gc->_nodes)
            this_gc->_curNode = this_gc->_nodes - 1;
        return D;
    }
    return (void *)0;
}

void * gc_rmvBackup(gc * this_gc)
{
    void * D;
    unsigned n;

    _gc_assert_valid(this_gc);
    if ((D = gc_atRmv(this_gc,n=this_gc->_curNode)) != (void *)0)
        if (n)
            this_gc->_curNode = n - 1;
        else
            this_gc->_curNode = this_gc->_nodes;
    return D;
}

int  gc_delBackup(gc * this_gc)
{
    void * D;

    _gc_assert_valid(this_gc);
    if (this_gc->_flags & GC_DEL)
        if ((D = gc_rmvBackup(this_gc)) != (void *)0)  {
            if (this_gc->vfT && this_gc->vfT->deleteD)
                (*(this_gc->vfT->deleteD))(this_gc,D);
            else
                GC_DELETE_DATA(D);
            return 1;
        }
    return 0;
}

int gc_clrBackup(gc * this_gc)
{
    _gc_assert_valid(this_gc);
    if (this_gc->_flags & GC_DEL)
        return gc_delBackup(this_gc);
    return (gc_rmvBackup(this_gc) != 0);
}

void * gc_next(gc * this_gc)
{
    _gc_assert_valid(this_gc);
    if (this_gc->linkS)  {
        if (this_gc->_curNode >= this_gc->_nodes)
            this_gc->_curNode = 0U;
        else
            this_gc->_curNode++;
        if (this_gc->_curNode < this_gc->_nodes)
            return this_gc->linkS[(this_gc->first+this_gc->_curNode)
                %this_gc->_limit];
    }
    return (void *)0;
}

void * gc_nextAsg(gc * this_gc, void * D)
{
    _gc_assert_valid(this_gc);
    if (this_gc->_curNode >= this_gc->_nodes)
        this_gc->_curNode = 0;  /*  wrap */
    else
        this_gc->_curNode++;
    if (gc_atGetAsg(this_gc,this_gc->_curNode,D))
        return D;
    /*  if this_gc->_curNode is a valid node, backup */
    if (!this_gc->_curNode)
        this_gc->_curNode = this_gc->_nodes;
    else if (this_gc->_curNode < this_gc->_nodes)
        this_gc->_curNode--;
    return 0;
}

void * gc_prev(gc * this_gc)
{
    _gc_assert_valid(this_gc);
    if (this_gc->linkS)  {
        if (this_gc->_curNode)  {
            if (this_gc->_curNode > this_gc->_nodes)
                this_gc->_curNode = this_gc->_nodes;
            this_gc->_curNode--;
        }
        else
            this_gc->_curNode = this_gc->_nodes;
        if (this_gc->_curNode < this_gc->_nodes)
            return this_gc->linkS[(this_gc->first+this_gc->_curNode)
                %this_gc->_limit];
    }
    return (void *)0;
}

void *  gc_prevAsg(gc * this_gc, void * D)
{
    _gc_assert_valid(this_gc);
    if (this_gc->_curNode)
    {
        this_gc->_curNode--;
        if (gc_atGetAsg(this_gc,this_gc->_curNode,D))
            return D;
        this_gc->_curNode++;
    }
    else
        this_gc->_curNode = this_gc->_nodes;
    return 0;
}

void gc_setCmP_specs(gc * this_gc, voidCmP cmP_, int resetSortFlag)
{
    _gc_assert_valid(this_gc);
    this_gc->_cmP = cmP_;
    if (resetSortFlag && this_gc->_nodes)
        this_gc->_flags &= ~GC_SORTED;
}

int   gc_sort_specs(gc * this_gc, voidCmP cmP_)
{
    unsigned low, mid, high;
    unsigned d;
    void * D;

/*
    The current node is always reset to undefined
    regardless of the outcome of sort.
*/
    _gc_assert_valid(this_gc);
    this_gc->_curNode = this_gc->_nodes;
    if (this_gc->_flags & GC_SORTED)  {
        if (this_gc->_cmP == cmP_ || !cmP_)
            return 1;
    }
    else if (!this_gc->_cmP && !cmP_)
        return 0;
    if (cmP_)  {
        this_gc->_cmP = cmP_;
        this_gc->_flags &= ~GC_SORTED;
    }
    if (!this_gc->_nodes || !this_gc->linkS)
        return (int) (this_gc->_flags |= GC_SORTED);
    if (this_gc->first)  {
        /* form contiguous block at front */
        d = (this_gc->first + this_gc->_nodes) % this_gc->_limit;
        if (d > this_gc->first)
            memmove(this_gc->linkS,&this_gc->linkS[this_gc->first],
                sizeof(void*)*this_gc->_nodes);
        else if (d < this_gc->first)
            memmove(&this_gc->linkS[d],
                &this_gc->linkS[this_gc->first],
                sizeof(void*)
                *(this_gc->_limit-this_gc->first));
        /* else array is full/contiguous */
        this_gc->first = 0U;
    }
    for (high = d = 1; d < this_gc->_nodes; high = ++d)  {
        low = 0U;
        D = this_gc->linkS[d];
        while (low < high)  {
            mid = ((high + low) >> 1);
            if ((*(this_gc->_cmP))(D,this_gc->linkS[mid]) <= 0)
                high = mid;
            else
                low = mid + 1;
        }
        if (high < d)  {
            memmove(&this_gc->linkS[high+1],
                &this_gc->linkS[high],
                sizeof(void*)*(d-high));
            this_gc->linkS[high] = D;
        }
    }
    return (int) (this_gc->_flags |= GC_SORTED);
}

void * gc_insSort(gc * this_gc, void * D)
{
    unsigned low, high;
    void * ok;

/*
    The current node is left undefined if
    anything fails, otherwise it is set to the
    newly inserted node.
*/
    _gc_assert_valid(this_gc);
    this_gc->_curNode = this_gc->_nodes;
    if (!this_gc->_cmP)
        return (void *)0;
    if (!D || this_gc->_nodes >= this_gc->_maxNodes)
        return (void *)0;
    if (!(this_gc->_flags & GC_SORTED))
        if (!gc_sort(this_gc))
            return (void *)0;
    low = 0U;
    high = this_gc->_nodes;  /*  this_gc->linkS == 0 -> nodes == 0! */
    while (low < high)  {
        this_gc->_curNode = ((high + low) >> 1);
        if ((*(this_gc->_cmP))(D,
            this_gc->linkS[(this_gc->first+this_gc->_curNode)
            %this_gc->_limit]) <= 0)
            high = this_gc->_curNode;
        else
            low = this_gc->_curNode + 1;
    }
    if ((ok = gc_atIns(this_gc,high,D)) == (void *)0)
        this_gc->_curNode = this_gc->_nodes;
    else
        this_gc->_curNode = high;
    /*  gc_atIns() resets sorted to zero  */
    this_gc->_flags |= GC_SORTED;
    return ok;
}

void * gc_insUnique(gc * this_gc, void * D)
{
    if (!gc_findFirst(this_gc,D))
        if (gc_sorted(this_gc))
            return gc_insSort(this_gc,D);
        else
            return gc_insQ(this_gc,D);
    return (void *)0;
}

void * gc_insUniqueNew(gc * this_gc, const void * D)
{
    if (!gc_findFirst(this_gc,D))
        if (gc_sorted(this_gc))
            return gc_insSortNew(this_gc,D);
        else
            return gc_insQNew(this_gc,D);
    return (void *)0;
}

void * gc_findFirst_specs
    (const gc * this_gc, const void * K, unsigned * idx,
    voidCmP kcmP, int honorSortFlag)
{
    unsigned low, high, _idx;

/*
    idx is left undefined if anything fails,
    otherwise it is set to the newly found node.
*/
    _gc_assert_valid(this_gc);
    *idx = _idx = this_gc->_nodes;
    if (!kcmP)  {
        honorSortFlag = 1;
        if ((kcmP = this_gc->_cmP) == voidCmP0)
            return (void *) 0;
    }
    if (!this_gc->linkS || !K || !this_gc->_nodes)
        return (void *)0;
    if (honorSortFlag && (this_gc->_flags & GC_SORTED))  {
        /*  binary search for this_gc->first match  */
        low = 0U;
        high = this_gc->_nodes - 1;
        while (low < high)  {
            _idx = ((high + low) >> 1);
            if ((*kcmP)(K,this_gc->linkS[(this_gc->first+_idx)
                %this_gc->_limit]) <= 0)
                high = _idx;
            else
                low = _idx + 1;
        }
        _idx = high;
        if ((*kcmP)(K,this_gc->linkS[(this_gc->first+high)
            %this_gc->_limit]))
            _idx = this_gc->_nodes;
    }
    else  {  /*  linear search!  */
        _idx = 0;
        do  {
            if ((*kcmP)(K,this_gc->linkS
                [(this_gc->first+_idx)
                %this_gc->_limit]))
                break;
        } while (++_idx < this_gc->_nodes);
    }
    *idx = _idx;
    return ((_idx < this_gc->_nodes)?
        this_gc->linkS[(this_gc->first+_idx)%this_gc->_limit]
        : (void *)0);
}

void * gc_findNext_specs
    (const gc * this_gc, const void * K, unsigned * idx,
    voidCmP kcmP, int honorSortFlag)
{
/*
    For sorted gc's you must first call
    gc_findFirst() to insure consistent results!
*/
    unsigned _idx;
    void * D;

/*
        idx is left undefined if anything fails,
        otherwise it is set to the newly found node.
*/
    _gc_assert_valid(this_gc);
    if (!kcmP)  {
        honorSortFlag = 1;
        kcmP = this_gc->_cmP;
    }
    if (!this_gc->linkS || !K || !kcmP  || !this_gc->_nodes
        || (++(*idx) == this_gc->_nodes))  {
        *idx = this_gc->_nodes;
        return (void *)0;
    }
    if (*idx > this_gc->_nodes)
        *idx = 0;
    if (!(*kcmP)(K,D=this_gc->linkS[(this_gc->first+*idx)
        %this_gc->_limit]))  {
        return D;
    }
    if (honorSortFlag && (this_gc->_flags & GC_SORTED))
        return (void *) 0;
    _idx = *idx;
    while (++_idx < this_gc->_nodes)
        if (!(*kcmP)(K,this_gc->linkS[(this_gc->first+_idx)
            %this_gc->_limit]))
            break;
    *idx = _idx;
    return ((_idx < this_gc->_nodes)?
        this_gc->linkS[(this_gc->first+_idx)%this_gc->_limit]
          : (void *)0);
}
