/*
        gc.h -- C Generic Container 2.0
        (C) Copyright 1996  John Webster Small
        All rights reserved
*/

#ifndef gc_h
#define gc_h

/* LINTLIBRARY */  /* Unix lint library module */

/* turn off PC-lint warning - boolean constant: _gc_assert_valid() */
/*lint -e506 */
/* turn off PC-lint warning - unusual pointer cast: _gc_assert_valid() */
/*lint -e740 */

/* heap manager hooks */

#define GC_NEW_HEADER  ((gc *)malloc(sizeof(gc)))
#define GC_DELETE_HEADER(this_gc) free(this_gc)
#define GC_NEW_VECTOR(nobj) ((void **)malloc(sizeof(void *)*nobj))
#define GC_DELETE_VECTOR(V) free(V)
#define GC_NEW_DATA(size) malloc(size)
#define GC_DELETE_DATA(D) free(D)

#ifndef limits_h
#include <limits.h>        /* UINT_MAX */
#ifndef limits_h
#define limits_h
#endif
#endif

#ifndef string_h
#include <string.h>        /* memcpy(), etc. */
#ifndef string_h
#define string_h
#endif
#endif

#ifndef assert_h
#include <assert.h>
#ifndef assert_h
#define assert_h
#endif
#endif

#ifndef stdlib_h
#include <stdlib.h>        /* malloc(), etc. */
#ifndef stdlib_h
#define stdlib_h
#endif
#endif


/*  gc compare pointer types  */

typedef int(*voidCmP)(const void * D1, const void * D2);
#define voidCmP0 ((voidCmP)0)

/*  gc constants  */

#define GC_MAXNODES     ((unsigned)(UINT_MAX/sizeof(void *)))
#define GC_LIMIT        20U
#define GC_DELTA        10U
#define GC_NOTFOUND     GC_MAXNODES
#define GC_SORTED       0x0001U
#define GC_BIND_ONLY    0x0000U
#define GC_ASG          0x0002U
#define GC_READ         0x0004U
#define GC_READ_DEL     0x0008U
#define GC_NEW          0x0010U
#define GC_NEWED        0x0020U
#define GC_DEL          0x0040U
#define GC_ALL_FLAGS    0x007FU
#define GC_AND         (GC_ASG|GC_NEW|GC_DEL)


/*  gc structures  */

typedef struct gc_header gc;

/* gc virtual function types */

typedef void * (* gc_vF_assignD )(const gc * this_gc, void * D,
   const void * S, unsigned flags);
typedef void * (* gc_vF_newD    )(const gc * this_gc, const void * D);
typedef void   (* gc_vF_deleteD )(const gc * this_gc, void * D);
typedef int    (* gc_vF_attachD )(const gc * this_gc, void * D);
typedef void   (* gc_vF_detachD )(const gc * this_gc, void * D);

/* gc virtual function table */

typedef struct
{
  gc_vF_assignD assignD;
  gc_vF_newD newD;
  gc_vF_deleteD deleteD;
  gc_vF_attachD attachD;
  gc_vF_detachD detachD;
} gc_vft;

/* gc virtual function (table) macros */

#define gc_vfT0 ((gc_vft *)0)

#define gc_vf_name(TYPE,FNC)  TYPE ## _ ## FNC
#define gc_vf_assignD(TYPE) \
  TYPE * gc_vf_name(TYPE,assignD)(const gc * this_gc, TYPE * D, \
    const TYPE * S, unsigned flags)
#define gc_vf_newD(TYPE) \
  TYPE * gc_vf_name(TYPE,newD)(const gc * this_gc, const TYPE * D)
#define gc_vf_deleteD(TYPE) \
  void gc_vf_name(TYPE,deleteD)(const gc * this_gc, TYPE * D)
#define gc_vf_attachD(TYPE) \
  int gc_vf_name(TYPE,attachD)(const gc * this_gc, TYPE * D)
#define gc_vf_detachD(TYPE) \
  void gc_vf_name(TYPE,detachD)(const gc * this_gc, TYPE * D)

#define gc_vft_name(TYPE) gc_vft_ ## TYPE
#define gc_vft_decl(TYPE) extern gc_vft gc_vft_name(TYPE)
#define gc_vft_def(TYPE,ASSIGN,NEW,DELETE,ATTACH,DETACH) \
  gc_vft gc_vft_name(TYPE) = { ASSIGN,NEW,DELETE,ATTACH,DETACH }

/* clone data for gc_vf_newD(TYPE) */
extern void * gc_memdup(const void * D, size_t size);

#define gc_decl(TYPE) gc_vft_decl(TYPE)
#define gc_def(TYPE) \
  static gc_vf_assignD(TYPE) \
    { this_gc = this_gc++; flags = flags++; \
      return (TYPE *) memcpy(D,S,sizeof(TYPE)); } \
  static gc_vf_newD(TYPE) \
    { this_gc = this_gc++; \
      return (TYPE *) gc_memdup(D,sizeof(TYPE)); } \
  gc_vft_def(TYPE, \
    (gc_vF_assignD)(gc_vf_name(TYPE,assignD)), \
    (gc_vF_newD)(gc_vf_name(TYPE,newD)), \
    (gc_vF_deleteD)(0), \
    (gc_vF_attachD)(0), \
    (gc_vF_detachD)(0))
#define gc_destruct_def(TYPE) \
  static gc_vf_assignD(TYPE) \
    { this_gc = this_gc++; flags = flags++; \
      return (TYPE *) memcpy(D,S,sizeof(TYPE)); } \
  static gc_vf_newD(TYPE) \
    { this_gc = this_gc++; \
      return (TYPE *) gc_memdup(D,sizeof(TYPE)); } \
  static gc_vf_deleteD(TYPE) \
    { this_gc = this_gc++; TYPE ## _destruct(D); } \
  gc_vft_def(TYPE, \
    (gc_vF_assignD)(gc_vf_name(TYPE,assignD)), \
    (gc_vF_newD)(gc_vf_name(TYPE,newD)), \
    (gc_vF_deleteD)(gc_vf_name(TYPE,deleteD)), \
    (gc_vF_attachD)(0), \
    (gc_vF_detachD)(0))
#define gc_delete_def(TYPE) \
  static gc_vf_assignD(TYPE) \
    { this_gc = this_gc++; flags = flags++; \
      return (TYPE *) memcpy(D,S,sizeof(TYPE)); } \
  static gc_vf_newD(TYPE) \
    { this_gc = this_gc++; \
      return (TYPE *) gc_memdup(D,sizeof(TYPE)); } \
  static gc_vf_deleteD(TYPE) \
    { this_gc = this_gc++; TYPE ## _delete(D); } \
  gc_vft_def(TYPE, \
    (gc_vF_assignD)(gc_vf_name(TYPE,assignD)), \
    (gc_vF_newD)(gc_vf_name(TYPE,newD)), \
    (gc_vF_deleteD)(gc_vf_name(TYPE,deleteD)), \
    (gc_vF_attachD)(0), \
    (gc_vF_detachD)(0))
#define gc_vfT(TYPE) & gc_vft_name(TYPE)

/* declare virtual functions (tables) for native types */
typedef char _gc_string;
gc_decl(_gc_string);
gc_decl(int);
gc_decl(long);
gc_decl(short);
gc_decl(unsigned);
typedef unsigned long _gc_ulong;
gc_decl(_gc_ulong);
typedef unsigned short _gc_ushort;
gc_decl(_gc_ushort);
gc_decl(float);
gc_decl(double);
typedef long double _gc_ldouble;
gc_decl(_gc_ldouble);

struct gc_header
{
        unsigned lowLimit, lowThreshold, first;
        void **  linkS;
        unsigned _limit,    _delta,   _nodes;
        unsigned _maxNodes, _flags, _curNode;
        voidCmP  _cmP;
        gc_vft * vfT;
        unsigned short dynamic;
};


/*  Constructors, Destructor, and Misc.  */

/* force constructors to dynamically allocate gc headers */
#define NEW_GC ((gc *)0)

extern gc * gc_init_specs
    (gc * this_gc_0, gc_vft * vfT, unsigned flags_);
#define gc_init(this_gc_0) \
    gc_init_specs(this_gc_0,gc_vfT0,GC_BIND_ONLY)
#define gc_init_type_specs(this_gc_0,TYPE,flags) \
    gc_init_specs(this_gc_0,&gc_vft_name(TYPE),flags)
#define gc_init_type(this_gc_0,TYPE) \
    gc_init_type_specs(this_gc_0,TYPE,GC_AND)

/* constructors for native types */
#define gc_string(this_gc_0) gc_init_type(this_gc_0,_gc_string)
#define gc_int(this_gc_0) gc_init_type(this_gc_0,int)
#define gc_long(this_gc_0) gc_init_type(this_gc_0,long)
#define gc_short(this_gc_0) gc_init_type(this_gc_0,short)
#define gc_unsigned(this_gc_0) gc_init_type(this_gc_0,unsigned)
#define gc_ulong(this_gc_0) gc_init_type(this_gc_0,_gc_ulong)
#define gc_ushort(this_gc_0) gc_init_type(this_gc_0,_gc_ushort)
#define gc_float(this_gc_0) gc_init_type(this_gc_0,float)
#define gc_double(this_gc_0) gc_init_type(this_gc_0,double)
#define gc_ldouble(this_gc_0) gc_init_type(this_gc_0,_gc_ldouble)

extern gc * gc_copyInit_specs
    (gc * this_gc_0, const gc * src_gc, int deep);
#define gc_copyInit(this_gc_0,src_gc)  \
    gc_copyInit_specs(this_gc_0,src_gc,1)
extern gc * gc_destruct(gc * this_gc);

/* misc. methods */
extern int gc_cmp(const gc * this_gc, const gc * b);


/*  Private Primitives: do not call!  */

#ifdef NDEBUG
#define GC_NDEBUG
#endif
#ifdef GC_NDEBUG
#define _gc_assert(X)
#define _gc_assert_valid(this_gc)
#define _gc_valid(this_gc) (this_gc)
#else
#define _gc_assert(X) assert(X)
#define _gc_assert_valid(this_gc)  \
     assert((this_gc) && (((void **)(this_gc)) != (this_gc)->linkS))
  extern gc * _gc_valid(const gc * this_gc);
#endif

extern void * _gc_atDeleteLeak
    (gc * this_gc, unsigned n, void * D,
     void * (*method)(gc *, unsigned, void *));
extern void * _gc_deleteLeak
    (gc * this_gc, void * D, void * (*method)(gc *, void *));
extern void * _gc_cloneD
    (gc * this_gc, const void  * D);

/*  Housekeeping Primitives  */

#define gc_limit(this_gc) ((const unsigned)(_gc_valid(this_gc)->_limit))
extern void gc_setLimit(gc * this_gc, unsigned newLimit);
#define gc_pack(this_gc)  \
     gc_setLimit(this_gc,(_gc_valid(this->gc)->_nodes))
#define gc_delta(this_gc) ((const unsigned)(_gc_valid(this_gc)->_delta))
extern void gc_setDelta(gc * this_gc, unsigned newDelta);
#define gc_nodes(this_gc) ((const unsigned)(_gc_valid(this_gc)->_nodes))
#define gc_maxNodes(this_gc) ((const unsigned)(_gc_valid(this_gc)->_maxNodes))
extern void gc_setMaxNodes(gc * this_gc, unsigned newMaxNodes);
#define gc_vacancy(this_gc)  \
    (_gc_valid(this_gc)->_maxNodes - _gc_valid(this_gc)->_nodes)
#define gc_vacancyNonElastic(this_gc)  \
    (_gc_valid(this_gc)->_limit - _gc_valid(this_gc)->_nodes)
#define gc_flags_specs(this_gc,flags_)  \
    (_gc_valid(this_gc)->_flags & (flags_))
#define gc_flags(this_gc) gc_flags_specs(this_gc,GC_ALL_FLAGS)
#define gc_setFlags(this_gc,flags_)  \
    ((void)(_gc_valid(this_gc)->_flags |= (flags_)))
#define gc_resetFlags(this_gc,flags_)  \
    ((void)(_gc_valid(this_gc)->_flags &= ~(flags_)))

/*  Elastic Array Primitives  */

extern void * gc_atIns(gc * this_gc, unsigned n, void * D);
#define gc_atInsDL(this_gc,n,D) _gc_atDeleteLeak(this_gc,n,D,gc_atIns)
#define gc_atInsNew(this_gc,n,D) gc_atInsDL(this_gc,n,_gc_cloneD(this_gc,D))
extern void * gc_atRmv(gc * this_gc, unsigned n);
extern void gc_allRmv(gc * this_gc);
extern int gc_atDel(gc * this_gc,unsigned n);
#define gc_atDelAsg(this_gc,n,D)  \
  (gc_atGetAsg(this_gc,n,D)?((void)gc_atDel(this_gc,n), D) : 0)
extern int gc_allDel(gc * this_gc);
#define atClr(this_gc,n)  \
  (gc_atDel(this_gc,n)? 1 : (gc_atRmv(this_gc,n) != (void *)0))
extern void gc_allClr(gc * this_gc);
extern void * gc_atPut(gc * this_gc, unsigned n, void * D);
extern void * gc_atPutNew(gc * this_gc, unsigned n, const void * D);
#define gc_atPutDL(this_gc,n,D) _gc_atDeleteLeak(this_gc,n,D,gc_atPut)
extern void * gc_atPutAsg(gc * this_gc, unsigned n, const void * D);
extern void * gc_atGet(const gc * this_gc, unsigned n);
extern void * gc_atGetAsg(gc * this_gc, unsigned n, void * D);
extern void * gc_atXchg(gc * this_gc, unsigned n, void * D);
extern unsigned gc_indexOf(const gc * this_gc, const void * D);


/*  Stack - Deque - Queue Primitives  */

#define gc_push(this_gc,D) gc_atIns(this_gc,0U,D)
#define gc_pushNew(this_gc,D) gc_atInsNew(this_gc,0U,D)
#define gc_pushDL(this_gc,D) gc_atInsDL(this_gc,0U,D)
#define gc_pop(this_gc) gc_atRmv(this_gc,0U)
#define gc_popDel(this_gc) gc_atDel(this_gc,0U)
#define gc_popDelAsg(this_gc,D) gc_atDelAsg(this_gc,0U,D)
#define gc_top(this_gc) gc_atGet(this_gc,0U)
#define gc_topAsg(this_gc,D) gc_atGetAsg(this_gc,0U,D)
#define gc_insQ(this_gc,D) gc_atIns(this_gc,gc_nodes(this_gc),D)
#define gc_insQNew(this_gc,D) gc_atInsNew(this_gc,gc_nodes(this_gc),D)
#define gc_insQDL(this_gc,D) gc_atInsDL(this_gc,gc_nodes(this_gc),D)
#define gc_unQ(this_gc) gc_atRmv(this_gc,gc_nodes(this_gc)-1)
#define gc_unQDel(this_gc) gc_atDel(this_gc,gc_nodes(this_gc)-1)
#define gc_unQDelAsg(this_gc,D) gc_atDelAsg(this_gc,gc_nodes(this_gc),D)
#define gc_rear(this_gc) gc_atGet(this_gc,gc_nodes(this_gc)-1)
#define gc_rearAsg(this_gc,D) gc_atGetAsg(this_gc,gc_nodes(this_gc)-1,D)


/*  List (single and double linked) Primitives  */

#define gc_curNode(this_gc)  \
  ((const unsigned)((_gc_valid(this_gc)->_curNode \
   < _gc_valid(this_gc)->_nodes)? \
   _gc_valid(this_gc)->_curNode : GC_NOTFOUND))
extern void * gc_setCurNode(gc * this_gc, unsigned n);
#define gc_resetCurNode(this_gc) gc_setCurNode(this_gc,GC_NOTFOUND)
extern void * gc_insBefore(gc * this_gc, void * D);
#define gc_insBeforeDL(this_gc,D) _gc_deleteLeak(this_gc,D,gc_insBefore)
#define gc_insBeforeNew(this_gc,D) \
    gc_insBeforeDL(this_gc,_gc_cloneD(this_gc,D))
extern void * gc_rmvStayPut(gc * this_gc);
extern int gc_delStayPut(gc * this_gc);
#define gc_delAsgStayPut(this_gc,D)  \
  (gc_atGetAsg(this_gc,gc_curNode(this_gc),D)?  \
   ((void)gc_delStayPut(this_gc), D) : 0)
extern int gc_clrStayPut(gc * this_gc);
extern void * gc_insAfter(gc * this_gc, void * D);
#define gc_insAfterDL(this_gc,D) _gc_deleteLeak(this_gc,D,gc_insAfter)
#define gc_insAfterNew(this_gc,D) \
    gc_insAfterDL(this_gc,_gc_cloneD(this_gc,D))
extern void * gc_rmvBackup(gc * this_gc);
extern int gc_delBackup(gc * this_gc);
#define gc_delAsgBackup(this_gc,D)  \
  (gc_atGetAsg(this_gc,gc_curNode(this_gc),D)?  \
   ((void)gc_delBackup(this_gc), D) : 0)
extern int gc_clrBackup(gc * this_gc);
#define gc_put(this_gc,D) gc_atPut(this_gc,gc_curNode(this_gc),D)
#define gc_putNew(this_gc,D) gc_atPutNew(this_gc,gc_curNode(this_gc),D)
#define gc_putDL(this_gc,D) gc_atPutDL(this_gc,gc_curNode(this_gc),D)
#define gc_putAsg(this_gc,D) gc_atPutAsg(this_gc,gc_curNode(this_gc),D)
#define gc_get(this_gc) gc_atGet(this_gc,gc_curNode(this_gc))
#define gc_getAsg(this_gc,D) gc_atGetAsg(this_gc,gc_curNode(this_gc),D)
extern void * gc_next(gc * this_gc);
extern void * gc_nextAsg(gc * this_gc, void * D);
extern void * gc_prev(gc * this_gc);
extern void * gc_prevAsg(gc * this_gc, void * D);
#define gc_forEach(this_gc)  \
    for ((void)gc_setCurNode((this_gc),0); \
         gc_get(this_gc); (void)gc_next(this_gc))
#define gc_do(this_gc,FNC) \
    gc_forEach(this_gc) (*FNC)(gc_get(this_gc))
#define gc_select(dst_gc,src_gc,SEL) \
    gc_forEach(src_gc)  \
       if ((*SEL)(gc_get(src_gc)))  \
           gc_insQ(dst_gc,gc_get(src_gc)); \
       else  /* null stmt */
#define gc_reject(dst_gc,src_gc,SEL) \
    gc_forEach(src_gc)  \
       if (!(*SEL)(gc_get(src_gc)))  \
           gc_insQ(dst_gc,gc_get(src_gc)); \
       else  /* null stmt */
#define gc_findAll(dst_gc,src_gc,K)  \
    if (gc_findFirst(src_gc,K))  \
      do { gc_insQ(dst_gc,gc_get(src_gc); }  \
      while gc_findNext(src_gc,K); \
    else  /* null stmt */
#define gc_collect(dst_gc,src_gc,COL) \
    gc_forEach(src_gc)  \
       gc_insQ(dst_gc,(*COL)(gc_get(src_gc)))


/* Priority Q, Set, Bag, Dictionary, Sort Primitives */

#define gc_sorted(this_gc) (_gc_valid(this_gc)->_flags & GC_SORTED)
#define gc_unSort(this_gc) \
 ((void)(_gc_valid(this_gc)->_flags &= ~GC_SORTED))
extern void gc_setCmP_specs(gc * this_gc, voidCmP cmP_, int resetSortFlag);
#define gc_setCmP(this_gc,cmP_) gc_setCmP_specs(this_gc,cmP_,1)
#define gc_resetCmP(this_gc) gc_setCmP_specs(this_gc,voidCmP0,1)
#define gc_cmP(this_gc) ((const voidCmP)(_gc_valid(this_gc)->_cmP))
extern int gc_sort_specs(gc * this_gc, voidCmP cmP_);
#define gc_sort(this_gc) gc_sort_specs(this_gc,voidCmP0)
extern void * gc_insSort(gc * this_gc, void * D);
#define gc_insSortDL(this_gc,D) _gc_deleteLeak(this_gc,D,gc_insSort)
#define gc_insSortNew(this_gc,D) \
    gc_insSortDL(this_gc,_gc_cloneD(this_gc,D))
extern void * gc_insUnique(gc * this_gc, void * D);
#define gc_insUniqueDL(this_gc,D) _gc_deleteLeak(this_gc,D,gc_insUnique)
extern void * gc_insUniqueNew(gc * this_gc, const void * D);

extern void * gc_findFirst_specs
  (const gc * this_gc, const void * K,
   unsigned * idx, voidCmP kcmP, int honorSortFlag);
#define gc_findFirst(this_gc,K) \
  gc_findFirst_specs(this_gc,K,&(_gc_valid(this_gc)->_curNode),voidCmP0,1)
extern void * gc_findNext_specs
  (const gc * this_gc, const void * K,
   unsigned * idx, voidCmP kcmP, int honorSortFlag);
#define gc_findNext(this_gc,K) \
  gc_findNext_specs(this_gc,K,(gc_assert(this_gc), \
  &(_gc_valid(this_gc)->_curNode)),voidCmP0,1)

#endif  /*  gc_h  */
